package flow.components
import chisel3._
import chisel3.util.experimental.decode.{decoder, TruthTable}
import chisel3.util.{Valid, log2Ceil}
import chisel3.util.MuxLookup
import chisel3.util.BitPat
import shapeless.{HNil, ::}
import shapeless.HList
import flow.Params
import RV32InstSubfields._
import flow.components.util._

object PcControlInterface {
  object SrcSelect extends ChiselEnum {
    val pStatic, pJmp, pBR = Value
  }

  class newPcControlInterface extends Bundle {
    val srcSelect = Input(SrcSelect())

    def ctrlBindPorts = {
      srcSelect :: HNil
    }
  }
  def apply() = new newPcControlInterface;
}

class ProgramCounter(implicit p: Params) extends Module {
  val control = IO(PcControlInterface())
  import PcControlInterface.SrcSelect._
  val in = IO(new Bundle {
    val brOffset = Input(UInt(p.XLEN))
    val jAddr = Input(UInt(p.XLEN))
  })
  val out = IO(Output(UInt(p.XLEN)))

  private val pcReg = RegInit(p.resetVector.U)
  out := pcReg

  private val npc = MuxLookup(control.srcSelect, 4.U)(
    Seq(
      pStatic -> (pcReg + 4.U),
      pJmp -> in.jAddr,
      pBR -> (pcReg + in.brOffset)
    )
  )

  pcReg := npc
}

class PcController(implicit p: Params) extends Module {
  val in = IO(new Bundle {
    val inst = Input(UInt(p.instWidth))
  })

  val out = IO(Flipped(PcControlInterface()))

  import RV32Inst._
  import PcControlInterface.SrcSelect._
  private val _jmpMapping = Array(jal, jalr).map(_ -> pJmp.BP)
  private val _brMapping =
    Array(beq, bne, blt, bge, bltu, bgeu).map(_ -> pBR.BP)

  val mapping = TruthTable(
    _jmpMapping ++ _brMapping,
    pStatic.BP
  )

  out.srcSelect := PcControlInterface.SrcSelect
    .safe(
      (
        decoder(in.inst, mapping)
      )
    )
    ._1
}
